@@ -10,6 +10,7 @@ option add *Button.highlightThickness 	0	startup
 option add *Entry.relief 		flat	startup
 option add *Entry.highlightThickness 	0	startup
 option add *font 			fixed 	startup
+option add *Text.c_link			blue	startup
 
 proc install_init { appName dotFile } {
     global install
@@ -83,6 +84,11 @@ proc install_version { var version {comment {}} } {
     set install(versionVar) $var
     install_var $var $version $comment
 }
+proc install_name { name value comment } {
+    global install
+    set install(namevar) $name
+    install_var $name $value $comment
+}
 proc install_dir { name value comment } {
     global install
     lappend install(dirlist) $name
@@ -177,11 +183,11 @@ proc installDoField { item {override 0} } {
     global install
     if ![info exists install(wuid)] { set install(wuid) 0 }
     incr install(wuid)
-    set f [frame .rim.import$install(wuid) -relief raised]
+    set f [frame .fields.import$install(wuid) -relief raised]
     set var [installFieldVar $item]
     button $f.label -text [format "%-30s:" [installFieldComment $item]] \
 	-command "installShowValue $var" -font fixed
-    entry $f.entry -font fixed -width 35
+    entry $f.entry -font fixed -width 40
     $f.entry insert 0 [installFieldDefault $item $override]
     bind $f.entry <Return> [list installSetValue $var]
     if {$install(lastentry) != {}} {
@@ -196,6 +202,7 @@ proc installDoField { item {override 0} } {
     pack $f -side top -expand true -fill both
     pack $f.label -side left -padx 3
     pack $f.entry -side right -expand true -fill both
+    return $f
 }
 proc installUpdateField { item {override 0} } {
     global install
@@ -348,7 +355,7 @@ proc installVerify {} {
 }
 
 proc installSed { } {
-    global install
+    global install program
     set id 0
     while {[catch {open /tmp/sed.$id w} script]} {
 	incr id
@@ -392,23 +399,27 @@ proc installSed { } {
     }
     puts $script ""
     close $script
+    set name [installGetValue $install(namevar)]
     foreach prog $install(sedProgs) {
+	regsub $program $prog $name newfile
 	if [catch {
-	    exec sed -f /tmp/sed.$id < ${prog}$install(sedSuffix) > $prog
-	    exec chmod a+x $prog
+	    exec sed -f /tmp/sed.$id < ${prog}$install(sedSuffix) > $newfile
+	    exec chmod a+x $newfile
 	} msg] {
 	    installFeedback "sed error on $prog: $msg"
 	}
     }
-#    exec rm /tmp/sed.$id
+    exec rm /tmp/sed.$id
 }
 proc installPatch {} {
-    global install
+    global install program
     installSave
     installVerify
     installSed
     set sample [lindex $install(sedProgs) 0]
-    catch {exec diff -c ${sample}$install(sedSuffix) $sample} diff
+    set name [installGetValue $install(namevar)]
+    regsub $program $sample $name name
+    catch {exec diff -c ${sample}$install(sedSuffix) $name} diff
     catch {destroy .test}
     set numLines [llength [split $diff \n]]
     if {$numLines == 0} {
@@ -452,8 +463,11 @@ proc installTest {} {
 	}
     }
     if [info exists install(test)] {
-	installFeedback $install(test)
-	eval $install(test)
+	set name [installGetValue $install(namevar)]
+
+	regsub exmh $install(test) $name test
+	installFeedback $test
+	eval $test
     } else {
 	installFeedback "No install_test command"
     }
@@ -511,7 +525,7 @@ proc installCmd { logProc unixCmd } {
     }
 }
 proc installInner { {logProc nolog} } {
-    global install
+    global install program
     installVerify
     installSed
     foreach dirType $install(dirlist) {
@@ -525,10 +539,23 @@ proc installInner { {logProc nolog} } {
 	    continue
 	}
 	if [info exists install(glob,$dirType)] {
+	    #
+	    # Tweak program name
+	    #
+	    if {$dirType == "bin"} {
+		set newglob {}
+		set name [installGetValue $install(namevar)]
+		foreach pat $install(glob,bin) {
+		    regsub $program $pat $name pat
+		    lappend newglob $pat
+		}
+	    } else {
+		set newglob $install(glob,$dirType)
+	    }
 	    #
 	    # Install glob pattern - copy the files in
 	    #
-	    foreach f [eval glob $install(glob,$dirType)] {
+	    foreach f [eval glob $newglob] {
 		if [catch {
 		    set t [file tail $f]
 		    if {$dirType == "man"} {
@@ -539,8 +566,13 @@ proc installInner { {logProc nolog} } {
 		    } else {
 			set newf $t
 		    }
-		    installCmd $logProc [list rm -f $dir/$newf]
-		    installCmd $logProc [list cp $f $dir/$newf]
+		    if [file isdirectory $f] {
+			installCmd $logProc [list rm -rf $dir/$newf]
+			installCmd $logProc [list cp -r $f $dir/$newf]
+		    } else {
+			installCmd $logProc [list rm -f $dir/$newf]
+			installCmd $logProc [list cp $f $dir/$newf]
+		    }
 		    if {$dirType == "bin"} {
 			installCmd $logProc [list chmod a+rx $dir/$newf]
 		    } else {
@@ -581,10 +613,11 @@ proc MakeDir { logProc dir } {
 proc installFake {} {
     global exmh install
     toplevel .fake
+    wm title .fake "Install Actions"
     frame .fake.top
     button .fake.top.quit -text "Dismiss" -command {destroy .fake}
     label .fake.top.label -text "  Pending install actions"
-    pack .fake.top -side top -fill both -expand true
+    pack .fake.top -side top -fill x
     pack .fake.top.quit -side left
     pack .fake.top.label -side left -fill both
 
@@ -604,23 +637,19 @@ proc install_dialog {} {
     wm title . "Exmh Install"
     wm minsize . 200 200
 
-    toplevel .info
-    wm title .info "Install info for $install(appName)"
-
-
-    text .info.t
-    .info.t config -yscrollcommand {.info.s set} -setgrid true
-    wm minsize .info 40 10
-    scrollbar .info.s -orient vert -command {.info.t yview}
-    pack .info.s -side right -fill y
-    pack .info.t -side left -fill both -expand true
-    .info.t insert 1.0 $install(helpText)
     update idletasks
 
     installBindInit
-    
-    frame .rim -bd 5 -relief flat
-    pack .rim -side top -expand true -fill both
+
+    frame .x
+    canvas .can -yscrollcommand ".x.scroll set"
+    scrollbar .x.scroll -command ".can yview"
+    pack .x.scroll -side right -fill y -padx 2 -pady 5
+    pack .can -in .x -side top -expand true -fill both -padx 5 -pady 5
+    pack .x -side top -expand true -fill both
+
+    frame .fields -relief flat
+    .can create window 0 0 -anchor nw -window .fields
 
     installFieldInit
     foreach v $install(sequence) {
@@ -629,9 +658,19 @@ proc install_dialog {} {
 	} else {
 	    set override 0
 	}
-	installDoField $install(field,$v) $override
+	set f [installDoField $install(field,$v) $override]
     }
     installFieldDone
+    tkwait visibility $f
+    set bbox [.can bbox all]
+    set dy [winfo height $f]
+    .can config -scrollregion $bbox \
+	-width [winfo width .fields] \
+	-height [expr 10 * $dy] \
+	-yscrollincrement $dy
+
+    frame .rim -bd 5 -relief flat
+    pack .rim -side top -fill x
 
     set install(msg) [label .rim.feedback -text "" -anchor w -padx 10]
     pack $install(msg) -side top -expand true -fill both
@@ -652,6 +691,7 @@ proc install_dialog {} {
     button .rim.buttons.readme -text "I have read the instructions" \
 	-command install_pack_buttons
     pack .rim.buttons.readme -side left
+    pack .rim.buttons.readme -side left
     pack .rim.buttons.quit -side right
 }
 proc install_pack_buttons {} {
@@ -858,3 +898,15 @@ proc installBindEntryInner { what } {
     }
 
 }
+proc install_html {html} {
+    global install program auto_path exwin widgetText
+    lappend auto_path ./lib
+
+    option add *Text*background white
+    option add *Text*foreground black
+    option add *Text*highlightBackground white
+    option add *Text*highlightColor black
+    proc Exmh_Status {string} {installFeedback $string}
+    set widgetText(smoothScroll) 0
+    set exwin(mtext) [Html_Display $html file:[file join [pwd] [info script]]]
+}
