@@ -129,11 +129,9 @@ proc FolderDelCommit { top entry } {
 	return
     }
     if {[file tail $nextid] == 1} {
-	catch {exec rm -f $path/$name/.xmhcache}
-	catch {
-	    foreach nfsjunk [glob $path/$name/.nfs*] {
-		exec rm -f $nfsjunk
-	    }
+	File_Delete $path/$name/.xmhcache
+	foreach nfsjunk [glob -nocomplain $path/$name/.nfs*] {
+	    File_Delete $nfsjunk
 	}
 	if [catch {exec rmf +$name -nointeractive} err] {
 	    Exmh_Status $err
