@@ -15,15 +15,15 @@ proc Ftoc_Init {} {
     global ftoc
     set ftoc(displayValid) 0		;# 1 => pick results, not full scan
     set ftoc(displayDirty) 0		;# 1 => display differs from cache
-    if ![info exists ftoc(skipMarked)] {
-	set ftoc(skipMarked) 1		;# Skip marked in Next, Prev
-    }
     set ftoc(mono) [string match mono* [tk colormodel .]]
     # Parameters to the Next button
     Preferences_Add "Scan Listing" \
 "These settings affect the behavior of Exmh as you move through the scan listing to view and mark messages.
 While the default for Auto Commit is OFF, I suggest you try it out.
 Messages are still temporarily marked, but the commit is done when you need it." {
+    {exwin(ftextLines)	ftextLines 15	{Scan listing lines}
+"Lines in the Scan listing window, which is
+also called Folder-Table-Of-Contents (FTOC)."}
     {ftoc(implied) impliedDirection ON "Implied Direction"
 "If set, Exmh will remember your current direction,
 next or previous, and go that way after you mark a
@@ -37,6 +37,35 @@ folder with unseen messages."}
 "If set, Exmh will invoke the Commit operation to
 commit deletions and refiles when it would otherwise
 just complain that such a commit is required."}
+    {ftoc(commitDialog) commitDialog ON "Commit Dialog"
+"If set, you get a confirmation dialog when exmh wants
+you to commit pending changes.  Otherwise you just
+get a warning message and have to hit the Commit button."}
+    {ftoc(autoPack) autoPack OFF "Auto Pack"
+"If set, Exmh will pack the folder every time a commit is performed."}
+    {ftoc(autoSort) autoSort OFF "Auto Sort"
+"If set, Exmh will sort the folder every time you change into it"}
+    {ftoc(autoSortType) autoSortType {CHOICE date subject sender custom} {Sorting criterion}
+"Sort by Date:/Subject:/From: or user-defined fields. Uses MH sortm command."}
+    {ftoc(autoSortCrit) autoSortCrit {-textfield keywords} {Custom criterion}
+"Custom parameters for sortm"}
+    {ftoc(showNew) ftocShowNew OFF "Show New Messages"
+"If set, Exmh will scroll the FTOC display to show
+new message that arrive because of an Inc."}
+    {ftoc(linkAdvance) advanceOnLink OFF "Advance after Link"
+"If set, Exmh will advance to the next message after a link."}
+    {ftoc(skipMarked) skipMarked ON "Next/Prev skip marked msgs"
+"If set, Next and Prev will skip over messages that have
+been marked for move or delete."}
+    {flist(cycleBack)	cycleBack ON	"Cycle back to first"
+"If there are no folders with unseen messages, then this
+option causes you to change to the first folder given by your
+Folder-Order MH profile entry."}
+    {ftoc(scanWidth) scanWidth 100 "Default scan width"
+"This value is passed as the -width argument to scan and in."}
+    {ftoc(scanSize) scanSize 100 "Default amount to scan"
+"Only the last N messages are scanned when you first enter a folder.
+The number is controlled by this setting."}
     }
 }
 proc Ftoc_Reset { numMsgs msgid folder } {
@@ -53,7 +82,7 @@ proc Ftoc_Reset { numMsgs msgid folder } {
     if {$msgid == {}} {
 	set ftoc(curLine) {}		;# current display line number
     } else {
-	set ftoc(curLine) [Ftoc_FindMsg $msgid]
+	set ftoc(curLine) {}		;# Set later in Msg_Change ?
     }
 }
 proc Ftoc_Update { numMsgs folder } {
@@ -65,16 +94,21 @@ proc Ftoc_Update { numMsgs folder } {
 
 proc Ftoc_Bindings { w } {
     # Bindings for the ftoc text widget
-    Bindings_Main $w
+    global tk_version
+    if {$tk_version >= 4.0} {
+	# The TScroll binding to too general.
+	# We'll do our own scroll bindings here.
+	bindtags $w [list $w]
+    }
 
     # Button-1 starts selection range
     bind $w <Button-1> {
 	FtocRangeStart [lindex [split [%W index current] .] 0]
-	focus %W
+	Exmh_Focus
     }
     bind $w <Shift-Button-1> {
 	FtocRangeAdd [lindex [split [%W index current] .] 0]
-	focus %W
+	Exmh_Focus
     }
     bind $w <B1-Motion> {
 	FtocRangeExtendXY %x %y
@@ -83,18 +117,23 @@ proc Ftoc_Bindings { w } {
 	FtocRangeExtendXY %x %y
     }
     bind $w <Any-ButtonRelease-1> {
-	FtocRangeEnd [lindex [split [%W index current] .] 0]
+	FtocRangeEnd [lindex [split [%W index current] .] 0] 0
     }
     bind $w <Shift-ButtonRelease-1> {
-	FtocRangeEnd [lindex [split [%W index current] .] 0] true
+	FtocRangeEnd [lindex [split [%W index current] .] 0] 1
     }
     bind $w <Button-3> {
 	set lineNumber [lindex [split [%W index current] .] 0]
 	Msg_Pick $lineNumber noshow
-	focus %W
+	Exmh_Focus
     }
     bind $w <Double-Button-1> { }
     bind $w <Triple-Button-1> { }
+
+    bind $w <Button-2> {WidgetTextMark %W %y}
+    bind $w <B2-Motion> {WidgetTextDragto %W %y $exwin(scrollSpeed)}
+
+    Drag_Attach $w FtocDragSelect Shift 3
 }
 proc FtocRangeStart { line } {
     # For normal button-down "start a selection"
@@ -115,41 +154,43 @@ proc FtocRangeAdd { line } {
     set ftoc(extend) 0
     FtocRangeInvert $line $line
 }
-proc FtocRangeEnd { {line {}} {addcurrent false} } {
+proc FtocRangeEnd { {line {}} {addcurrent 0} } {
     # For end of button sweep
     global ftoc exwin
-    set ftoc(extend) 0
+    catch {unset ftoc(extend)}
     if ![info exists ftoc(pickend)] {
 	# Spurious button-release event
 	return
     }
+    if {($line == $ftoc(pickstart)) && !$addcurrent} {
+	# regular button click optimization
+	unset ftoc(pickend)
+	Msg_Pick $line show
+	return
+    }
     if {$line != {}} {
 	FtocRangeExtend $line
     }
-    set lineset {}
-    foreach range [concat \
-	[FtocMakePairs [$exwin(ftext) tag ranges range]] \
-	[FtocMakePairs [$exwin(ftext) tag ranges drange]] \
-	[FtocMakePairs [$exwin(ftext) tag ranges mrange]]] {
-	set mark1 [lindex $range 0]
-	set line [lindex [split $mark1 .] 0]
-	lappend lineset $line
-    }
-    FtocPickRange $lineset $addcurrent
-    unset ftoc(pickend)
+    FtocPickRange $addcurrent
+    catch {unset ftoc(pickend)}
 }
-proc Ftoc_PickMsgs { ids } {
+proc Ftoc_PickMsgs { ids addtosel } {
     # For adding to the selection by message number
     global ftoc
     Exmh_Status "Marking [llength $ids] hits"
     set lines {}
     for {set L 0} {$L <= $ftoc(numMsgs)} {incr L} {
-	if {[lsearch $ids [Ftoc_MsgNumber $L]] >= 0} {
+	set ix [lsearch $ids [Ftoc_MsgNumber $L]]
+	if {$ix >= 0} {
 	    lappend lines $L
+	    set ids [lreplace $ids $ix $ix]
+	    if {[string length $ids] == 0} {
+		break
+	    }
 	}
     }
     Ftoc_LinesHighlight $lines
-    FtocPickRange $lines
+    FtocPickRange $addtosel
 }
 proc FtocRangeExtendXY { x y } {
     global ftoc exwin widgetText
@@ -182,19 +223,21 @@ proc FtocRangeExtendXY { x y } {
 proc FtocSelExtend {} {
     global ftoc exwin widgetText
     set w $exwin(ftext)
-    if {$ftoc(extend) != 0} {
-	catch {
-	    set delta [expr {$ftoc(extend) / 16}]
-	    if {$delta == 0} {
-		set delta [expr { ($ftoc(extend) < 0) ? -1 : 1 }]
-	    }
-	    set newmark [expr {$ftoc(lastmark) + $delta}]
-	    FtocRangeExtend $newmark
-	    set ftoc(lastmark) $newmark
-	    $w yview -pickplace $newmark.0
-	    after $widgetText(selectDelay) [list FtocSelExtend]
+    if {![info exists ftoc(extend)] ||
+	($ftoc(extend) == 0)} {
+	return
+    }
+    catch {
+	set delta [expr {$ftoc(extend) / 16}]
+	if {$delta == 0} {
+	    set delta [expr { ($ftoc(extend) < 0) ? -1 : 1 }]
 	}
-    }    
+	set newmark [expr {$ftoc(lastmark) + $delta}]
+	FtocRangeExtend $newmark
+	set ftoc(lastmark) $newmark
+	$w yview -pickplace $newmark.0
+	after $widgetText(selectDelay) [list FtocSelExtend]
+    }
 }
 proc FtocRangeExtend { line } {
     global ftoc
@@ -300,13 +343,22 @@ proc FtocRangeInvert { start end } {
 	catch {
 	    set newtag range
 	    set oldtag {}
+	    set nuke 0
 	    foreach tag [$win tag names $line.0] {
 		case $tag {
 		    deleted { set newtag drange ; set oldtag $tag ; break; }
 		    moved { set newtag mrange ; set oldtag $tag ; break; }
 		    range { set newtag {} ; set oldtag $tag ; break; }
-		    drange { set newtag deleted ; set oldtag $tag ; break; }
-		    mrange { set newtag moved ; set oldtag $tag ; break; }
+		    drange { set newtag deleted ; set oldtag $tag ;
+			    set nuke 1; break}
+		    mrange { set newtag moved ; set oldtag $tag ;
+			    set nuke 1; break; }
+		}
+	    }
+	    if {$nuke} {
+		set ix [lsearch $ftoc(lineset) $line]
+		if {$ix >= 0} {
+		    set ftoc(lineset) [lreplace $ftoc(lineset) $ix $ix]
 		}
 	    }
 	    if {$oldtag != {}} {
@@ -423,34 +475,48 @@ proc Ftoc_FindMsg { msgid {line {}} } {
     if {$msgid == {}} {
 	return {}
     }
-    set start $ftoc(numMsgs)
-    set lastMsg [Ftoc_MsgNumber $start]
-    if {$lastMsg <= $msgid} {
-	# Search forward
-	for {set L $start} {$L <= $ftoc(numMsgs)} {incr L} {
-	    if {[Ftoc_MsgNumber $L] == $msgid} {
-		return $L
-	    }
+#
+# Linear search for pick and thread FTOCs (pseudo-displays)
+#
+    if !$ftoc(displayValid) {
+        for {set L 1} {$L <= $ftoc(numMsgs)} {incr L} {
+            if {[Ftoc_MsgNumber $L] == $msgid} {
+                return $L
+            }
+        }
+        return {}
+    }
+
+    set min 1
+    set max $ftoc(numMsgs)	;# Ignore trailing blank line
+    while (1) {
+	set m1 [Ftoc_MsgNumber $min]
+	if {$msgid == $m1} {
+	    return $min
 	}
-	for {set L 1} {$L < $start} {incr L} {
-	    if {[Ftoc_MsgNumber $L] == $msgid} {
-		return $L
-	    }
+	set m2 [Ftoc_MsgNumber $max]
+	if {$msgid == $m2} {
+	    return $max
 	}
-    } else {
-	# Search backward
-	for {set L $start} {$L > 0} {incr L -1} {
-	    if {[Ftoc_MsgNumber $L] == $msgid} {
-		return $L
-	    }
+	if {$msgid > $m2 || $msgid < $m1} {
+	    Exmh_Status "Cannot find $msgid ($m1,$m2)" warn
+	    return "" ;# new message not listed
 	}
-	for {set L $ftoc(numMsgs)} {$L >= $start} {incr L -1} {
-	    if {[Ftoc_MsgNumber $L] == $msgid} {
-		return $L
-	    }
+	if {$max == $min} {
+	    return ""	;# not found
+	}
+	set next [expr int(($max+$min)/2)]
+	set m3 [Ftoc_MsgNumber $next]
+	if {$m3 > $msgid} {
+	    set max $next
+	} elseif {$min == $next} {
+	    Exmh_Status "Cannot find $msgid" warn
+	    return "" ;# new message not listed
+	} else {
+	    set min $next
 	}
     }
-    return {}
+    # not reached
 }
 proc Ftoc_MsgNumber { L } {
     global exwin
@@ -466,20 +532,33 @@ proc Ftoc_MsgNumberRaw { line } {
 	return ""
     }
 }
-proc FtocPickRange { lineset {addcurrent false}} {
+proc FtocPickRange { {addcurrent 0} } {
     # Select a range of messages, or add to the current range
+    # Because of toggle/inverted selections, we pretty much
+    # have to recompute the select set from range tags
     global exwin ftoc
+    set lineset {}
+    if {$ftoc(curLine) != {}} {
+	if {$addcurrent} {
+	    Ftoc_RangeHighlight $ftoc(curLine) $ftoc(curLine)
+	}
+	Ftoc_ClearCurrent
+	Msg_ClearCurrent
+	set ftoc(curLine) {}
+    }
+    foreach range [concat \
+	[FtocMakePairs [$exwin(ftext) tag ranges range]] \
+	[FtocMakePairs [$exwin(ftext) tag ranges drange]] \
+	[FtocMakePairs [$exwin(ftext) tag ranges mrange]]] {
+	set mark1 [lindex $range 0]
+	set line [lindex [split $mark1 .] 0]
+	lappend lineset $line
+    }
     if {$lineset == {}} {
 	return			;# spurious <ButtonRelease-1> events
     }
     set ftoc(lineset) $lineset
     set ftoc(pickone) 0
-    if {($addcurrent != "false") && ($ftoc(curLine) != {})} {
-	if {[lsearch $lineset $ftoc(curLine)] < 0} {
-	    lappend ftoc(lineset) $ftoc(curLine)
-	    Ftoc_RangeHighlight $ftoc(curLine) $ftoc(curLine)
-	}
-    }
     if {[llength $ftoc(lineset)] == 1} {
 	# This calls Msg_Change,
 	# which calls Ftoc_ClearCurrent, which sets pickone to 1,
@@ -487,14 +566,6 @@ proc FtocPickRange { lineset {addcurrent false}} {
 	Msg_Pick [lindex $ftoc(lineset) 0] show
     } else {
 	Buttons_Range	;# Enable actions on ranges
-	foreach line $ftoc(lineset) {
-	    if {$ftoc(curLine) == $line} {
-		# Unhighlight current if it is within range.
-		# This is a bit lame, but fixes display problems that
-		# stem from combining the current and range looks.
-		$exwin(ftext) tag remove current $ftoc(curLine).0 $ftoc(curLine).end
-	    }
-	}
     }
 }
 proc Ftoc_PickSize {} {
@@ -506,6 +577,21 @@ proc Ftoc_PickSize {} {
 	return $len
     }
 }
+proc Ftoc_NewFtoc {} {
+    global ftoc
+    set ids {}
+    foreach l $ftoc(lineset) {
+	lappend ids [Ftoc_MsgNumber $l]
+    }
+    if {[llength $ids] <= 1} {
+	Exmh_Status "Select more than one message first" warn
+	return
+    }
+    if {[Ftoc_Changes "new ftoc"] == 0} {
+	Exmh_Status $ids
+	Scan_ProjectSelection $ids
+    }
+}
 
 # Ftoc_ClearCurrent and Ftoc_Change are two parts of
 # dinking the ftoc display when advancing a message.
@@ -518,6 +604,7 @@ proc Ftoc_ClearCurrent {} {
 
     if {$ftoc(curLine) != {}} {
 	$exwin(ftext) tag remove current $ftoc(curLine).0 $ftoc(curLine).end
+	$exwin(ftext) tag remove currentBg $ftoc(curLine).0 $ftoc(curLine).end
 	Ftoc_RescanLine $ftoc(curLine)
     }
     return $ftoc(curLine)
@@ -533,29 +620,26 @@ proc Ftoc_Change { msgid line {show show} } {
 	}
 	Ftoc_RescanLine $ftoc(curLine) +
 	$exwin(ftext) tag add current $ftoc(curLine).0 $ftoc(curLine).end
-	FtocHackCurrentHighlight
-	$exwin(ftext) yview -pickplace [expr $ftoc(curLine)-1]
+	$exwin(ftext) tag add currentBg $ftoc(curLine).0 $ftoc(curLine).end
+	set top [$exwin(ftext) index @0,4]
+	if [catch {expr {$top+1}}] {set top 0}	;# trap 100.-1 format, iconic
+	if {$ftoc(curLine) == $top ||
+	    $ftoc(curLine) == $top+$exwin(ftextLines)-1} {
+	    WidgetTextYview $exwin(ftext) [expr $ftoc(curLine)-$exwin(ftextLines)/2].0
+	} else {
+	    WidgetTextYview $exwin(ftext) -pickplace $ftoc(curLine).0
+	}
 	set ok 1
     }
     return $ok
 }
-proc FtocHackCurrentHighlight {} {
-    global exwin ftoc
-    if {$ftoc(mono)} {
-	# Monochrome hack
-	if [regexp {(deleted|moved|range)} \
-		    [$exwin(ftext) tag names $ftoc(curLine).0]] {
-	    $exwin(ftext) tag remove current \
-		    $ftoc(curLine).0 $ftoc(curLine).end
-	}
-    }
-}
 proc Ftoc_ShowUnseen { folder } {
     global exwin flist
     set unseen [Flist_UnseenMsgs $folder]
     if {[llength $unseen] > 0} {
 	set end [$exwin(ftext) index end]
 	set line [lindex [split $end .] 0]
+	set msgNum 0
 	for {} {$line > 0} {incr line -1} {
 	    set msgNum [Ftoc_MsgNumber $line]
 	    set i [lsearch $unseen $msgNum]
@@ -567,21 +651,38 @@ proc Ftoc_ShowUnseen { folder } {
 		}
 	    }
 	}
+	# Repair bogus unseen sequences
+	# msgNum is the smallest message number, but it might not be
+	# the first message in the folder because of short scans
+	# Anything in the unseen sequence above msgNum is probably wrong
+	# and can result from races with the background process
+	foreach id $unseen {
+	    if {$id > $msgNum} {
+		Flist_MsgSeen $id
+	    }
+	}
     } else {
+	Flist_SeenAll $folder	;# clear highlighting
 	return 0
     }
 }
 proc Ftoc_MarkSeen { ids } {
-    global exwin
-    foreach msgid $ids {
-	set L [Ftoc_FindMsg $msgid]
-	if {$L != {}} {
+    global exwin ftoc
+
+    set ids [lsort -integer -decreasing $ids]
+    for {set L $ftoc(numMsgs)} {$L > 0} {incr L -1} {
+	set ix [lsearch $ids [Ftoc_MsgNumber $L]]
+	if {$ix >= 0} {
 	    $exwin(ftext) tag remove unseen $L.0 $L.end
+	    set ids [lreplace $ids $ix $ix]
+	    if {[llength $ids] == 0} {
+		break
+	    }
 	}
     }
 }
 proc Ftoc_RescanLine { ix {plus none} } {
-    global exmh exwin
+    global exmh exwin ftoc
     if [catch {
 	set text [$exwin(ftext) get ${ix}.0 ${ix}.end]
 	set ok 0
@@ -601,6 +702,10 @@ proc Ftoc_RescanLine { ix {plus none} } {
 		if ![regexp {( *)([0-9]+).-} $text] {
 		    set ok [regsub {( *[0-9]+.)(.)} $text {\1-} newtext]
 		}
+		# Annotations result in writes to the directory.
+		# Here we mark the display dirty to force an update
+		# of the cache and prevent later rescans.
+		set ftoc(displayDirty) 1
 	    }
 	}
 	if {$ok} {
@@ -617,21 +722,20 @@ proc Ftoc_RescanLine { ix {plus none} } {
 	Exmh_Error "FtocRescanLine $ix : $msg"
     }
 }
+proc Ftoc_NextImplied { {show show} {implied implied} } {
+    global ftoc
+    if {$ftoc(implied) && $ftoc(direction) == "prev"} {
+	Ftoc_Prev $show
+    } else {
+	Ftoc_Next $show $implied
+    }
+}
 proc Ftoc_Next { show {implied no} } {
     # Go to the next message in the scan display
     global exmh flist ftoc
 
-    if {$implied != "no" && $ftoc(implied)} {
-	if {$ftoc(direction) == "prev"} {
-	    Ftoc_Prev $show
-	    return
-	}
-    } else {
-	set ftoc(direction) "next"
-    }
-    Exmh_Debug Ftoc_Next curline=$ftoc(curLine)
+    set ftoc(direction) "next"
     if {$ftoc(curLine) == {}} {
-	# No current; go for the next unseen.
 	if [Msg_ShowUnseen] {
 	    return
 	}
@@ -641,46 +745,7 @@ proc Ftoc_Next { show {implied no} } {
 	($ftoc(curLine) >= $ftoc(numMsgs)) || \
 	($ftoc(curLine) <= 0)} {
 	# End of folder
-	# Try to chain to the next folder with unread messages.
-	if {$implied != "no"} {
-	    # Implied - chained with some other operation - be lenient
-	    if {$ftoc(changed) > 0} {
-		# Special case if last message is marked.
-		# Nuke current highlighting because it conflicts with
-		# current highlight on mono screen.  Better fix is to
-		# have separate look for deleted-current.
-		if {$ftoc(curLine) != {}} {
-		    FtocHackCurrentHighlight
-if {0} {
-		    global exwin
-		    if [regexp {(deleted|moved|range)} \
-				[$exwin(ftext) tag names $ftoc(curLine).0]] {
-			$exwin(ftext) tag remove current \
-				$ftoc(curLine).0 $ftoc(curLine).end
-		    }
-}
-		}
-		Exmh_Status ""
-		Exmh_Status "Changes pending; End of folder" blue
-		return
-	    }
-	}
-	foreach f [Flist_UnseenFolders] {
-	    if {$f != $exmh(folder)} {
-		if {$ftoc(softChange)} {
-		    Folder_Change $f
-		    return
-		} else {
-		    set ftoc(softChange) 1
-		    Msg_ClearCurrent
-		    Exmh_Status ""
-		    Exmh_Status "End of folder; <Next> => $f" blue
-		    return
-		}
-	    }
-	}
-	Exmh_Status ""
-	Exmh_Status "End of folder" blue
+	Ftoc_NextFolder $implied
     } else {
 	# Simple case - go to the next message.
 	Msg_Pick $next $show
@@ -699,8 +764,45 @@ proc Ftoc_Prev { {show show} } {
     if {$ftoc(curLine) > 1} then {
 	set ftoc(direction) "prev"
 	Msg_Pick [FtocSkipMarked $ftoc(curLine) -1] $show
+    } else {
+	Ftoc_Next $show implied
     }
 }
+proc Ftoc_NextFolder { {implied no} } {
+    global ftoc
+    # Try to chain to the next folder with unread messages.
+    if {$implied != "no"} {
+	# Implied - chained with some other operation - be lenient
+	if {$ftoc(changed) > 0} {
+	    # Dirty folder - do not change.
+	    # If on last message, clear display because the
+	    # message is moved or deleted
+	    if {$ftoc(curLine) != {}} {
+		Ftoc_ClearCurrent
+		Msg_ClearCurrent
+	    }
+	    Exmh_Status ""
+	    Exmh_Status "Changes pending; End of folder" warn
+	    return
+	}
+    }
+    set f [Flist_NextUnseen]
+    if {[string length $f] != 0} {
+	if {$ftoc(softChange)} {
+	    Folder_Change $f Msg_ShowUnseen
+	    return
+	} else {
+	    set ftoc(softChange) 1
+	    Ftoc_ClearCurrent
+	    Msg_ClearCurrent
+	    Exmh_Status ""
+	    Exmh_Status "End of folder; <Next> => $f" warn
+	    return
+	}
+    }
+    Exmh_Status ""
+    Exmh_Status "End of folder" warn
+}
 proc Ftoc_PrevMarked { {show show} } {
     global ftoc
     set skip $ftoc(skipMarked)
@@ -708,6 +810,23 @@ proc Ftoc_PrevMarked { {show show} } {
     Ftoc_Prev $show
     set ftoc(skipMarked) $skip
 }
+proc Ftoc_Marked { id } {
+    global ftoc exwin
+    if {$ftoc(skipMarked) == 0} {
+	return 0	;# Pretend it isn't marked
+    }
+    set i [Ftoc_FindMsg $id]
+    if {[string length $i] == 0} {
+	return 1	;# Can't find it, pretend it's marked
+    }
+    set marked 0
+    foreach tag [$exwin(ftext) tag names $i.0] {
+	if [regexp {(deleted|moved|drange|mrange)} $tag] {
+	    set marked 1 ; break;
+	}
+    }
+    return $marked
+}
 proc FtocSkipMarked {start inc} {
     global exwin ftoc
 
@@ -735,21 +854,63 @@ proc Ftoc_Changes {type {allowAuto 1} } {
     global ftoc
 
     if {$ftoc(changed) != 0} then {
+	Exmh_Debug Ftoc_Changes $type
 	if {("$allowAuto" == "1") && $ftoc(autoCommit)} {
-	    Folder_Commit
+	    Folder_CommitType $type
 	    return 0
 	}
 	if {$type != {}} {
-	    Exmh_Status "$ftoc(changed) Changes pending: $type cancelled" red
+	    if {[string compare $type iconified] == 0} {
+		set msg "$ftoc(changed) changes pending"
+	    } else {
+		if {$ftoc(commitDialog) &&
+		    [FtocDialog $ftoc(changed) $type]} {
+		    Folder_CommitType $type
+		    Exmh_Focus
+		    return 0
+		} else {
+		    set msg "$ftoc(changed) changes pending: $type cancelled"
+		}
+	    }
+	    Exmh_Focus
+	    Exmh_Status $msg warn
 	    Sound_Error
 	} else {
-	    Exmh_Status "Oops, $ftoc(changed) left over changes" red
+	    Exmh_Status "Oops, $ftoc(changed) left over changes" error
 	    set ftoc(changed) 0
 	    return 1
 	}
+    } else {
+	Msg_CheckPoint		;# Sync unseen and cur message state.
     }
     return $ftoc(changed)
 }
+proc FtocDialog { changes type } {
+    global exwin ftoc
+    if [winfo exists $exwin(mtext).commit] {
+	destroy $exwin(mtext).commit
+    }
+    set f [frame $exwin(mtext).commit -class Dialog -bd 4 -relief ridge]
+    set blurb [expr {($changes > 1) ? "are $changes changes" : "is one change"}]
+    Widget_Message $f msg -text \
+"There $blurb pending.
+(Press Return to Commit)
+(Press <Control-c> to Cancel)" -aspect 1000
+    set but [Widget_Frame $f but Dialog {top expand fill} -bd 10]
+    set ftoc(okToCommit) 0
+    Widget_AddBut $but cancel "Cancel" {set ftoc(okToCommit) 0}
+    Widget_AddBut $but ok "Commit and $type" {set ftoc(okToCommit) 1}
+    focus $but
+    bind $but <Return> "$but.ok flash ; $but.ok invoke"
+    bind $but <Control-c> "$but.cancel flash ; $but.cancel invoke"
+    Widget_PlaceDialog $exwin(mtext) $exwin(mtext).commit
+    Visibility_Wait $but
+    catch {grab $but}
+    tkwait variable ftoc(okToCommit)
+    catch {grab release $but}
+    destroy $exwin(mtext).commit
+    return $ftoc(okToCommit)
+}
 
 proc Ftoc_Iterate { lineVar body } {
     global ftoc
@@ -772,26 +933,28 @@ proc Ftoc_Unmark {} {
 
     set hits 0
     Ftoc_Iterate line {
-	if [FtocUnmarkInner $line] { incr hits -1}
+	if [FtocUnmarkInner $line] { incr hits }
     }
     Exmh_Status "Unmarked $hits msgs"
-    incr ftoc(changed) $hits
+    incr ftoc(changed) -$hits
 }
-proc FtocUnmarkInner { line } {
+proc FtocUnmarkInner { line {all 0}} {
     global exwin
     set res 0
+    if {$all} {
+	set pat (deleted|moved|drange|mrange|copied)
+    } else {
+	set pat (deleted|moved|drange|mrange)
+    }
     foreach tag [$exwin(ftext) tag names $line.0] {
-	if [regexp {(deleted|moved|drange|mrange)} $tag] {
+	if [regexp $pat $tag] {
 	    $exwin(ftext) tag remove $tag $line.0 $line.end
-	    if [regexp {(drange|mrange)} $tag] {
+	    if [regexp {(drange|mrange|crange)} $tag] {
 		eval $exwin(ftext) tag add range $line.0 $line.end
 	    }
 	    set res 1
 	}
     }
-    if {! $res} {
-	Exmh_Debug FtocUnmarkInner $line miss
-    }
     return $res
 }
 proc Ftoc_Delete { line msgid } {
@@ -804,7 +967,7 @@ proc Ftoc_Delete { line msgid } {
 proc Ftoc_RemoveMark { line msgid } {
     # Flag the current message(s) for deletion
     global ftoc exwin
-    if ![FtocUnmarkInner $line] {
+    if ![FtocUnmarkInner $line 1] {
 	incr ftoc(changed)
     }
 
@@ -830,12 +993,36 @@ proc Ftoc_MoveMark { line msgid } {
 	$exwin(ftext) tag add mrange $line.0 $line.end
     }
 }
-proc Ftoc_Commit { rmmCommit moveCommit } {
+proc Ftoc_CopyMark { line msgid } {
+    global ftoc exwin exmh
+    if ![FtocUnmarkInner $line] {
+	incr ftoc(changed)
+    }
+    # This tag records the target folder
+    $exwin(ftext) tag add [list copied $exmh(target)] $line.0 $line.end
+
+    if {$ftoc(pickone)} {
+	$exwin(ftext) tag add moved $line.0 $line.end
+    } else {
+	$exwin(ftext) tag remove range $line.0 $line.end
+	$exwin(ftext) tag add mrange $line.0 $line.end
+    }
+}
+proc Ftoc_Commit { rmmCommit moveCommit copyCommit } {
     global ftoc exwin
-    Exmh_Status "Committing $ftoc(changed) changes..." red
+
+    # Disable operations on ranges
+    Ftoc_RangeUnHighlight
+    if {! $ftoc(pickone)} {
+	Buttons_Range 0
+	set ftoc(lineset) {}
+	set ftoc(pickone) 1
+    }
+
+    Exmh_Status "Committing $ftoc(changed) changes..."
     $exwin(ftext) configure -state normal
     FtocCommit deleted $rmmCommit
-    FtocCommit moved $moveCommit
+    FtocCommit moved $moveCommit $copyCommit
     $exwin(ftext) configure -state disabled
     set l $ftoc(curLine)
     if {$l == {}} {
@@ -845,75 +1032,107 @@ proc Ftoc_Commit { rmmCommit moveCommit } {
 	WidgetTextYview $exwin(ftext) $l
     }
     if {! [Ftoc_Changes {} noautocommit]} {
-	Exmh_Status "ok" blue
+	Exmh_Status "ok"
     }
 }
-proc FtocCommit {tagname commitProc} {
-    global ftoc exmh exwin
+proc FtocCommit {tagname commitProc {copyCommitProc {}} } {
+    global ftoc exmh exwin msg
 
     set delmsgs {}
+    set curid [file tail $msg(path)]
     set pairs [FtocMakeReversePairs [$exwin(ftext) tag ranges $tagname]]
     foreach range $pairs {
 	set c0 [lindex $range 0]
 	set ce [lindex $range 1]
 	scan $c0 "%d" L
+	set msgid [Ftoc_MsgNumber $L]
 	set F {}
+	set delline 0	;# Nuke display line
 	foreach tag [$exwin(ftext) tag names $c0] {
 	    if {([llength $tag] == 2) && ([lindex $tag 0] == "moved")} {
 		set F [lindex $tag 1]
+		# Build up a list of moved messages
+		# Note that the original order of the messages is maintained,
+		# (We are going from bottom to top thru the display.)
+		# The scan lines are reversed, which is handled by Scan_Move.
 		if ![info exists movemsgs($F)] {
-		    set movemsgs($F) {}
+		    set movemsgs($F) $msgid
+		} else {
+		    set movemsgs($F) [concat $msgid " " $movemsgs($F)]
 		}
+		lappend movescan($F) [$exwin(ftext) get $c0 "$ce + 1 chars"]
+		set delline 1
+	    }
+	    if {([llength $tag] == 2) && ([lindex $tag 0] == "copied")} {
+		set F [lindex $tag 1]
+		# Build up a list of moved messages
+		# Note that the original order of the messages is maintained,
+		# (We are going from bottom to top thru the display.)
+		# The scan lines are reversed, which is handled by Scan_Move.
+		if ![info exists copymsgs($F)] {
+		    set copymsgs($F) $msgid
+		} else {
+		    set copymsgs($F) [concat $msgid " " $copymsgs($F)]
+		}
+		lappend movescan($F) [$exwin(ftext) get $c0 "$ce + 1 chars"]
 	    }
 	}
-	set msgid [Ftoc_MsgNumber $L]
 	if {$tagname == "deleted"} {
 	    # Batch up deletes
 	    lappend delmsgs $msgid
-	} else {
-	    # Build up a list of moved messages
-	    # Note that the original order of the messages is maintained,
-	    # (We are going from bottom to top thru the display.)
-	    # The scan lines are reversed, which is handled by Scan_Move.
-#	    set movemsgs($F) [concat $movemsgs($F) " " $msgid]
-	    set movemsgs($F) [concat $msgid " " $movemsgs($F)]
-	    lappend movescan($F) [$exwin(ftext) get $c0 "$ce + 1 chars"]
+	    set delline 1
 	}
-	Msg_UnSeen $msgid	;# avoid MH mark bug
 	Flist_MsgSeen $msgid	;# in case deleted or moved w/out viewing
-	$exwin(ftext) delete $c0 "$ce + 1 chars"
-	set ftoc(displayDirty) 1
-	if {$L == $ftoc(curLine)} {
-	    set ftoc(curLine) {}
-	    Msg_ClearCurrent
-	} else {
-	    if {$L < $ftoc(curLine)} {
-		incr ftoc(curLine) -1
-		if {$ftoc(curLine) == 0} {
-		    set ftoc(curLine) {}
+	if {$delline} {
+	    Msg_UnSeen $msgid	;# avoid MH mark bug
+	    $exwin(ftext) delete $c0 "$ce + 1 chars"
+	    set ftoc(displayDirty) 1
+	    if {$msgid == $curid} {
+		Ftoc_ClearCurrent
+		Msg_ClearCurrent
+	    }
+	    if {$L == $ftoc(curLine)} {
+		set ftoc(curLine) {}
+	    } elseif {$ftoc(curLine) != {}} {
+		if {$L < $ftoc(curLine)} {
+		    incr ftoc(curLine) -1
+		    if {$ftoc(curLine) == 0} {
+			set ftoc(curLine) {}
+		    }
 		}
 	    }
+	    incr ftoc(numMsgs) -1
+	} else {
+	    FtocUnmarkInner $L
 	}
-	incr ftoc(numMsgs) -1
 	incr ftoc(changed) -1
     }
     if {$delmsgs != {}} {
-	Exmh_Status "$commitProc $delmsgs" red
+	Exmh_Status "$commitProc $delmsgs"
 	if [catch {
-	    Background_Rmm $commitProc $exmh(folder) $delmsgs
+	    BgAction "Rmm $exmh(folder)" $commitProc $exmh(folder) $delmsgs
 	} err] {
-	    Exmh_Status $err purple
+	    Exmh_Status $err error
+	}
+    }
+    # Do copies before links so you can both move and copy a message.
+    if {[catch {array names copymsgs} flist] == 0} {
+	foreach f $flist {
+	    Exmh_Status "Copying to $f, $copymsgs($f)"
+	    if [catch {
+		BgAction "Refile $f" $copyCommitProc $exmh(folder) $copymsgs($f) $f
+	    } err] {
+		Exmh_Status $err error
+	    }
 	}
     }
     if {[catch {array names movemsgs} flist] == 0} {
 	foreach f $flist {
-	    Exmh_Status "Refiling to $f, $movemsgs($f)" red
-	    set ok [Scan_CacheValid $f]
-	    set new [file tail [Mh_Path $f new]]
+	    Exmh_Status "Refiling to $f, $movemsgs($f)"
 	    if [catch {
-		Background_Refile $commitProc $exmh(folder) $movemsgs($f) $f $ok $movescan($f) $new
+		BgAction "Refile $f" $commitProc $exmh(folder) $movemsgs($f) $f
 	    } err] {
-		Exmh_Status $err purple
+		Exmh_Status $err error
 	    }
 	}
     }
@@ -957,6 +1176,9 @@ proc Ftoc_MoveFeedback { msgid line } {
 	if [regexp {moved (.+)} $tag match folder] {
 	    Exmh_Status "$msgid => +$folder"
 	    return
+	} elseif [regexp deleted $tag] {
+	    Exmh_Status "$msgid Pending Delete"
+	    return
 	}
     }
     Exmh_Status $msg
@@ -967,6 +1189,33 @@ proc Ftoc_FindNext {} {
 proc Ftoc_FindPrev {} {
     Find_It back
 }
+proc Ftoc_FindAll {string} {
+    global exwin find ftoc
+    if {[string length $string] == 0} {
+	Exmh_Status "No search string" warn
+	return -1
+    }
+    set msgids {}
+    for {set L 1} 1 {incr L} {
+	if [$exwin(ftext) compare $L.end >= end] {
+	    break
+	}
+	if [catch {$exwin(ftext) get $L.0 $L.end} text] {
+	    break
+	}
+	if [regexp -nocase -- $string $text] {
+	    lappend msgids [Ftoc_MsgNumberRaw $text]
+	}
+    }
+    if {[llength $msgids] == 0} {
+	Exmh_Status "No match" warn
+	return 0
+    } else {
+	Ftoc_PickMsgs $msgids 0
+	return 1
+    }
+
+}
 proc Ftoc_FindMatch {L string} {
     global exwin ftoc
     if {$L == $ftoc(lasthit)} {
@@ -975,11 +1224,133 @@ proc Ftoc_FindMatch {L string} {
     if [catch {$exwin(ftext) get $L.0 $L.end} text] {
 	return -1	;# off the end or beginning
     }
-    if [regexp -nocase $string $text] {
+    if [regexp -nocase -- $string $text] {
 	set ftoc(lasthit) $L
 	Msg_Pick $L show
 	return 1
     }
     return 0
 }
+proc Ftoc_Yview {args} {
+    global exwin
+    eval {WidgetTextYview $exwin(ftext)} $args
+}
+proc Ftoc_Advance { advance } {
+    global ftoc
+    if {[string compare $advance "advance?"] == 0} {
+	return $ftoc(linkAdvance)
+    } else {
+	return $advance
+    }
+}
+proc Ftoc_PageUp {} {
+    global exwin
+    Widget_TextPageUp $exwin(ftext)
+}
+proc Ftoc_PageDown {} {
+    global exwin
+    Widget_TextPageDown $exwin(ftext)
+}
+
+proc Ftoc_Sort {} {
+    global ftoc
+    case $ftoc(autoSortType) {
+	{date} { Folder_Sort -datefield date }
+	{subject} { Folder_Sort -textfield subject }
+	{sender} { Folder_Sort -textfield from }
+	{custom} { eval Folder_Sort $ftoc(autoSortCrit) }
+    }
+}
+proc Ftoc_SelectAll {} {
+    global ftoc
+    FtocRangeStart 1
+    FtocRangeEnd $ftoc(numMsgs)
+}
+proc Ftoc_SelectAllToEnd {} {
+    global ftoc
+    if {$ftoc(curLine) != {}} {
+      FtocRangeStart $ftoc(curLine)
+    } else {
+      if {$ftoc(direction) == "next"} {
+        FtocRangeStart 1
+      } else {
+        FtocRangeStart $ftoc(numMsgs)
+      }
+    }
+    if {$ftoc(direction) == "next"} {
+      FtocRangeEnd $ftoc(numMsgs)
+    } else {
+      FtocRangeEnd 1
+    }
+}
+proc Ftoc_CatchUp {} {
+    global ftoc
+    Ftoc_SelectAll
+    Msg_Remove
+    Folder_Commit
+    Msg_PageOrNext
+}
+proc Ftoc_CatchUpToEnd {} {
+    global ftoc
+    Ftoc_SelectAllToEnd
+    Msg_Remove
+    Folder_Commit
+    Msg_PageOrNext
+}
+#
+# Interface to Drag & Drop
+#
+set ftocDrag(types) {foldermsg}
+set ftocDrag(formats) {string filename}
+set ftocDrag(format,foldermsg) string
+set ftocDrag(format,filename) string
+set ftocDrag(type,string) foldermsg
+
+# Drag Selected
+proc FtocDragSelectOld {w x y wx wy} {
+	global exmh ftoc
+
+        set folder $ftoc(folder)
+        if !$ftoc(displayValid) {
+            set folder $exmh(folder)
+        }
+	set line [lindex [split [$w index current] .] 0]
+	set msg [Ftoc_MsgNumber $line]
+	if {$msg == {} || $msg == 0} return
+
+	# Hand off to Drag code
+	global ftocDrag mhProfile
+	set ftocDrag(source) $w
+	set ftocDrag(data,foldermsg) "+$folder $msg"
+	set ftocDrag(data,filename) $mhProfile(path)/$folder/$msg
+	Drag_Source ftocDrag $x $y
+}
+proc FtocDragSelect {w x y wx wy} {
+    global exmh ftoc ftocDrag mhProfile
+
+    set folder $ftoc(folder)
+    if !$ftoc(displayValid) {
+        set folder $exmh(folder)
+    }
+    set msgs {}
+    if $ftoc(pickone) {
+	set line [lindex [split [$w index current] .] 0]
+	set msgs [Ftoc_MsgNumber $line]
+	if {$msgs == {} || $msgs == 0} return
+	set ftocDrag(data,filename) $mhProfile(path)/$folder/$msgs
+    } else {
+	foreach line $ftoc(lineset) {
+	    set msgid [Ftoc_MsgNumber $line]
+	    if {$msgid != {}} {
+		lappend msgs $msgid
+	    }
+	}
+    	catch {unset ftocDrag(data,filename)}
+    }
+
+    # Hand off to Drag code
+    set ftocDrag(source) $w
+    set ftocDrag(data,foldermsg) "+$folder $msgs"
+    Drag_Source ftocDrag $x $y
+  }
 
