@@ -16,7 +16,7 @@ proc Help_KeyDisplay {} {
     if [Exwin_Toplevel .key "Exmh Key" Key] {
 	set key .key
 	set b .key.but
-    
+
 	set l [Widget_Label $b]
 	$l configure -text "[tk colormodel .]"
 
@@ -25,7 +25,7 @@ proc Help_KeyDisplay {} {
 	set key [Widget_Frame .key rim Rim]
 	$key configure -borderwidth 10
     
-	set t [Widget_SimpleText $key t {top fillx} -width 32 -height 9 -font $font -wrap none]
+	set t [Widget_SimpleText $key t {top fillx} -width 34 -height 12 -font $font -wrap none]
 	Ftoc_ColorConfigure $t
 	$t configure -state normal
 	$t delete 0.0 end
@@ -34,7 +34,13 @@ proc Help_KeyDisplay {} {
 	$t tag add moved   2.0 2.end
 	$t tag add deleted 3.0 3.end
 	$t tag add unseen  4.0 4.end
-	$t insert insert "\nleft   => change folder\nmiddle => view subfolders\nright  => refile message(s)\nshift-right => link message(s)"
+	$t insert insert "\nFolder Label Bindings\n"
+	$t insert insert "left          => change folder\n"
+	$t insert insert "middle        => view subfolders\n"
+	$t insert insert "right         => refile message(s)\n"
+	$t insert insert "shift-right   => link message(s)\n"
+	$t insert insert "shift-middle  => drag folder\n"
+	$t insert insert "control-right => clear target"
 	$t configure -state disabled
     
 	set c [canvas $key.can -width 20 -height 30]
@@ -60,26 +66,42 @@ proc HelpKeyLabel { c x y text font texttag boxtag} {
     return $x
 }
 
-proc Help { {suffix {}} } {
+proc Help { {name {Intro}} {title {Help for exmh} }} {
     global exmh
-    if {[string length $suffix] == 0} {
-	set suffix help
+    if [string match /* $name] {
+	# absolute pathname
+	set dir [file dirname $name]
+	set file [file tail $name]
+	set id help$file
+	set topic $file
+    } else {
+	set dir $exmh(library)/html
+	set file $dir/$name.html
+	set id help$name
+	set topic $name
     }
-    set id [string tolower $suffix]
-    set label [string toupper [string index $suffix 0]][string range $suffix 1 end]
-    if [Exwin_Toplevel .$id "Exmh $suffix" Help] {
-	Widget_Label .$id.but label {left fill} -text "  $label file for exmh"
-    
-	set t [Widget_Text .$id 30 -setgrid true]
+    if [string match *.html $file] {
+	return [Html_Window file:$file]
+    }
+    regsub -all {\.} $id _ id
+    if [Exwin_Toplevel .$id "Exmh $topic" Help] {
+	Widget_Label .$id.but label {left fill} -text "  $title"
+
+	set t [Widget_Text .$id 25 -setgrid true]
+	bind $t <Key-Next> {%W yview scroll 1 page}
+	bind $t <Key-Prior> {%W yview scroll -1 page}
+	bind $t <Key-Home> {%W see 1.0}
+	bind $t <Key-End> {%W see end}
 	Ftoc_ColorConfigure $t
-	$t insert insert "EXMH Version: $exmh(version)\n\n"
-	foreach dir [list $exmh(library) .] {
-	    if [catch {open $dir/exmh.$suffix} in] {
+	$t insert insert "EXMH Version: $exmh(version)\n"
+	foreach d [list $dir .] {
+	    if [catch {open $d/$file} in] {
 		continue
 	    }
 	    $t insert insert [read $in]
-	    # This is data-dependent...
-	    if {$suffix == "help"} {
+	    close $in
+	    # This is data-dependent..., but cute
+	    if {$name == "Intro"} {
 		set L 26
 		$t tag add current $L.0 $L.end ; incr L
 		$t tag add deleted $L.0 $L.end ; incr L
@@ -87,11 +109,15 @@ proc Help { {suffix {}} } {
 		$t tag add unseen $L.0 $L.end ; incr L
 	    }
 	    $t configure -state disabled
+	    scan [$t index end] %d height
+	    if {$height < 25} {
+		$t config -height $height
+	    }
 	    return $t
 	}
-	$t insert insert "Cannot find $suffix file to display"
+	$t insert insert "Cannot find $file file to display"
 	$t configure -state disabled
     }
-    return {}
+    return .$id.t
 }
 
