@@ -74,6 +74,7 @@ setting in your ~/.mh_profile file to a new directory."
 	set col 3
 	set nframes [expr [llength $folders] / $col]
 	for {set i 0} {$i <= $nframes} {incr i} {
+	    catch {destroy $w.but$i}
 	    set f [Widget_Frame $w but$i]
 	    for {set j 0} {$j < $col} {incr j} {
 		set next [lindex $folders 0]
@@ -89,6 +90,7 @@ setting in your ~/.mh_profile file to a new directory."
 
 	$_import(win,okbut) config -text Import -command ImportIt
 	set f [winfo parent $_import(win,okbut)]
+	catch {destroy $f.clear}
 	Widget_AddBut $f clear "Unselect All" [list ImportUnselectAll $f.clear]
     }
 }
@@ -112,26 +114,27 @@ proc ImportIt {} {
     global _importlist tk_version
     set t [Help Import "Log of Import Actions"]
     $t config -state normal -height 20
+    $t tag configure fixed -font fixed
     foreach name [lsort [array names _importlist]] {
 	set f $_importlist($name)
 	if {[string length $f] != 0} {
+	    global import
 	    if [catch {glob $f} file] {
 		Exmh_Status $file
 		continue
 	    }
 	    Exmh_Status "inc +$name -file $file -notruncate"
-	    $t insert end "inc +$name -file $file -notruncate"
+	    $t insert end "inc +$name -file $file -notruncate" fixed
 	    update idletasks
-	    catch {
-		MhExec inc +$name -file $file -notruncate < /dev/null
-	    } result
-	    $t insert end $result
-	    if {$tk_version >= 4.0} {
-		$t see end
-	    } else {
-		$t yview -pickplace end
+	    if [catch {
+		set in [open "|inc +$name -file $file -notruncate < /dev/null"]
+		fileevent $in readable [list ImportRead $in $t $name]
+	    } result] {
+		$t insert end $result
+		catch {close $in}
 	    }
-	    update idletasks
+	    $t see end
+	    tkwait variable import($name)
 	}
     }
     $t config -state disabled
@@ -139,3 +142,19 @@ proc ImportIt {} {
     destroy .import
 }
 
+proc ImportRead {in t name} {
+    global import
+
+    if [eof $in] {
+	catch {close $in}
+	set import($name) complete
+	return
+    }
+    if [catch {gets $in line} err] {
+	$t insert end $err\n
+	set import($name) complete
+	catch {close $in}
+    }
+    $t insert end $line\n
+}
+
