@@ -169,8 +169,8 @@ proc MsgSeen { msgid } {
     if {[lsearch $msg(seen) $msgid] < 0} {
 	lappend msg(seen) $msgid
     }
-    Flag_MsgSeen
     Flist_MsgSeen $msgid
+    Flag_MsgSeen
 }
 proc Msg_UnSeen { msgid } {
     # We nuke deleted and moved messages from the seen list because
@@ -200,7 +200,6 @@ proc Msg_Seen {} {
 # not it is assumed to be arguments to the MH command.
 
 proc Msg_Compose { args } {
-    global exmh msg	;# allow args to include $exmh(folder) $msg(id)
     if {[string length $args] == 0} {
 	set args Mh_CompSetup
     }
@@ -211,11 +210,7 @@ proc Msg_Compose { args } {
 	    return
 	}
     } else {
-	if [catch {
-	    Exmh_Status "comp $args"
-	    eval {MhExec comp -nowhatnowproc} $args
-	} err] {
-	    Exmh_Status "comp: $err"
+	if {![eval MsgComp $args]} {
 	    return
 	}
     }
@@ -230,20 +225,77 @@ proc Msg_CompUse {folder id} {
 	Msg_Compose +$folder $id
     }
 }
+# Compose a message to a particular person
+proc Msg_CompTo {address args} {
+    global mhProfile
+    if {![eval MsgComp $args]} {
+	return
+    }
+    set draftID [Mh_Cur $mhProfile(draft-folder)]
+    if {$draftID == {}} {
+	return
+    }
+    # re-write draft with new To: header
+    if [catch {
+	set path $mhProfile(path)/$mhProfile(draft-folder)/$draftID
+	set in [open $path]
+	set X [read $in]
+	close $in
+	regsub -nocase "(^|\n)to:\[^\n\]*\n" $X "\\1To: $address\n" X
+	set out [open $path w]
+	puts -nonewline $out $X
+	close $out
+    } err] {
+	Exmh_Status $err
+	return
+    }
+    Edit_Draft					;# Run the editor
+}
+# Use current selection as To: header
+proc Msg_CompSel {args} {
+    global mhProfile
+    if {[catch {selection get} address]} {
+	Exmh_Status "Select an address for Msg_CompTo"
+	return
+    }
+    eval {Msg_CompTo $address} $args
+}
+
+# General wrapper around comp
+proc MsgComp {args} {
+    # allow args to include $exmh(folder) $msg(id) $mhProfile(path)
+    global exmh msg mhProfile	
+    set exmh(ctype) {comp}
+    if [catch {
+	set ix [lsearch $args -form]
+	if {$ix < 0} {
+	    if [file exists $mhProfile(path)/$exmh(folder)/components] {
+		lappend args -form $exmh(folder)/components
+	    }
+	}
+	Exmh_Status "comp $args"
+	eval {MhExec comp -nowhatnowproc} $args
+    } err] {
+	Exmh_Status "comp: $err"
+	return 0
+    }
+    return 1
+}
 proc Msg_Reply { args } {
-    global exmh msg
+    global exmh msg mhProfile
+    set exmh(ctype) {repl}
     if {[string length $args] == 0} {
 	set args Mh_ReplySetup
     }
 
     if [MsgOk $msg(id) m] {
-	Mh_AtLink $exmh(folder) $m			;# Symlink to @
+	Quote_MakeFile $exmh(folder) $m
 	set edit 1
 	if {[string compare [info command $args] $args] == 0} {
 	    # Old interface with hook procedure
 	    if [catch {$args $exmh(folder) $m} err] {	;# Setup draft msg
 		Exmh_Status "${args}: $err" purple
-		Mh_AtLinkCleanup				;# Nuke @ link
+		Quote_Cleanup
 		return
 	    }
 	} else {
@@ -254,11 +306,18 @@ proc Msg_Reply { args } {
 		    set edit 0
 		    set args [lreplace $args $ix $ix]
 		}
+		set ix [lsearch $args -form]
+		if {$ix < 0} {
+		    if [file exists $mhProfile(path)/$exmh(folder)/replcomps] {
+			lappend args -form $exmh(folder)/replcomps
+			Exmh_Status "repl $args" purple
+		    }
+		}
 		eval {MhExec repl +$exmh(folder) $m -nowhatnowproc} $args
 		MhAnnoSetup $exmh(folder) $m repl
 	    } err] {	;# Setup draft msg
 		Exmh_Status "repl: $err" purple
-		Mh_AtLinkCleanup				;# Nuke @ link
+		Quote_Cleanup				;# Nuke @ link
 		return
 	    }
 	}
@@ -272,6 +331,7 @@ proc Msg_Reply { args } {
 
 proc Msg_Forward { args } {
     global exmh msg
+    set exmh(ctype) {forw}
     if {[string length $args] == 0} {
 	set args Mh_ForwSetup
     }
@@ -303,6 +363,13 @@ proc Msg_Forward { args } {
 		if {[lsearch $args -mime] >= 0} {
 		    set mime 1
 		}
+		set ix [lsearch $args -form]
+		if {$ix < 0} {
+		    if [file exists $mhProfile(path)/$exmh(folder)/forwcomps] {
+			lappend args -form $exmh(folder)/forwcomps
+			Exmh_Status "forw +$exmh(folder) $ids $args"
+		    }
+		}
 		eval {MhExec forw +$exmh(folder)} $ids -nowhatnowproc $args
 		MhAnnoSetup $exmh(folder) $ids forw
 	    } err] {
@@ -321,6 +388,7 @@ proc Msg_Forward { args } {
 
 proc Msg_Dist { args } {
     global exmh msg
+    set exmh(ctype) {dist}
     if {[string length $args] == 0} {
 	set args Mh_DistSetup
     }
@@ -553,6 +621,7 @@ proc Msg_MarkUnseen {} {
     }
     Msg_ClearCurrent
     Ftoc_ClearCurrent
+    Flist_ForgetUnseen $exmh(folder)
     Ftoc_ShowUnseen $exmh(folder)
 }
 
@@ -584,6 +653,14 @@ proc Msg_LineUp {} {
     global exwin
     Widget_TextLineUp $exwin(mtext)
 }
+proc Msg_Top {} {
+    global exwin
+    Widget_TextTop $exwin(mtext)
+}
+proc Msg_Bottom {} {
+    global exwin
+    Widget_TextBottom $exwin(mtext)
+}
 proc Msg_CopySelection {} {
     global exwin sedit
     catch {set sedit(killbuf) [$exwin(mtext) get sel.first sel.last]}
