@@ -28,7 +28,7 @@ proc Report_Bug {} {
     Edit_DraftID [file tail $draft]
 }
 proc Report_Registration {} {
-    global mhProfile exmh env
+    global mhProfile exmh env tcl_version tcl_patchLevel
 
     set draft [Mh_Path $mhProfile(draft-folder) new]
     if [catch {open $draft w} out] {
@@ -36,21 +36,22 @@ proc Report_Registration {} {
 	return
     }
     puts $out \
-"To: Brent.Welch@eng.sun.com
+"To: welch@acm.org
 Subject: Register exmh user
 -----
 $exmh(version) $env(USER)
 [exec uname -a]
+Tcl $tcl_patchLevel
 
 >   Please register as an exmh user so I can more accurately
 >   track the usage of exmh.  I will not use your email address
->   for any purpose, unless you join one of the mailing lists
->   listed below.  Any and all comments are appreciated.
+>   for any purpose other than to display a cool gif in my
+>   exmh face canvas.  Any and all comments are appreciated.
 >
 >   If you have registered for an earlier release you need not
 >   register again, unless you want to, of course.
 >
->	Brent Welch - Sun Labs - Brent.Welch@eng.sun.com
+>	Brent Welch <welch@acm.org>
 
 
 Please comment on exmh:
@@ -70,8 +71,8 @@ proc Report_Subscribe {list what} {
 	Exmh_Status "Cannot write $draft"
 	return
     }
-    puts $out "To: majordomo@sunlabs.sun.com"
-    puts $out "Subject: exmh mailing list"
+    puts $out "To: $list-request@redhat.com"
+    puts $out "Subject: $what"
     puts $out "------"
     puts $out "$what $list"
     puts $out "--"
