@@ -25,7 +25,7 @@ proc SeditId { draft } {
 	return $newdraft
     }
 }
-proc SeditSigfileDefault {} {
+proc SeditProperSigfileDefault {} {
     global sedit
     if {[string length $sedit(sigfileDefault)] == 0} {
 	set sedit(sigfileDefault) ~/.signature
@@ -35,12 +35,23 @@ proc SeditSigfileDefault {} {
     }
     return [glob -nocomplain $sedit(sigfileDefault)]
 }
+proc SeditSigfileDefault {} {
+    global intelligentSign
+    set propersig [SeditProperSigfileDefault]
+    if {$intelligentSign(state)} {
+	return {///auto///}
+    }
+    return $propersig
+}
 proc Sedit_Start { draft } {
-    global sedit
+    global sedit intelligentSign quote
     global exmh	;# for menu references to $exmh(...)
     if ![info exists sedit(init)] {
 	Sedit_Init
     }
+    if ![info exists sedit(checkpoint)] {
+	SeditPeriodicSave
+    }
     set id [SeditId $draft]
     set b .sedit${id}.but
     if {[Exwin_Toplevel .sedit$id $draft Sedit] == 0} {
@@ -53,7 +64,8 @@ proc Sedit_Start { draft } {
     } else {
 	wm iconname .sedit$id draft/$id
 	set f [Widget_Frame .sedit$id f Frame {top expand fill}]
-	set t [Widget_Text $f $sedit(height) -setgrid true]
+	set t [Widget_Text $f $sedit(height) -setgrid true -wrap char]
+
 	Drop_Attach $t SeditDragDrop
 	set sedit($t,status) [Widget_Entry .sedit${id} status {top fill} -relief raised]
 
@@ -101,7 +113,8 @@ proc Sedit_Start { draft } {
 		$b.sign config -text $txt...
 	    }
 	    if {$sedit(autoSign)} {
-		Widget_RadioMenuItem $menu "(none)" { } sedit($t,sigfile) -value {}
+		Widget_RadioMenuItem $menu "(none)" { } sedit($t,sigfile) {}
+		Widget_RadioMenuItem $menu "(intelligent)" { } sedit($t,sigfile) {///auto///}
 		$menu add separator
 		set i 1
 	    } else {
@@ -110,7 +123,7 @@ proc Sedit_Start { draft } {
 	    foreach file $sigFiles {
 		if {$sedit(autoSign)} {
 		    incr i
-		    Widget_RadioMenuItem $menu [file tail $file] { } sedit($t,sigfile) -value $file
+		    Widget_RadioMenuItem $menu [file tail $file] { } sedit($t,sigfile) $file
 #		    if {[string compare [file tail $file] [file tail $sedit(sigfileDefault)]] == 0} {
 #			$menu invoke $i
 #		    }
@@ -157,15 +170,20 @@ proc Sedit_Start { draft } {
     Exwin_ToplevelFocus .sedit$id $t
 
     SeditTextBindings $draft $t		;# set up sendMsg binding
-    if [file readable "@"] {
+    if [file readable $quote(filename)] {
 	$b.repl configure -state normal -command \
-		[list SeditInsertFile $draft $t "@"]
+		[list SeditInsertFile $draft $t $quote(filename)]
     } else {
 	$b.repl configure -state disabled
     }
     set sedit($t,keep) $sedit(keepDefault)
-    set sedit($t,format) $sedit(formatDefault)
     set sedit($t,mhn) $sedit(mhnDefault)
+    set sedit($t,format) $sedit(formatChoice)
+    switch $sedit($t,format) {
+	OnType	{$t config -wrap char}
+	OnSend	{$t config -wrap word}
+	Never	{$t config -wrap none}
+    }
     switch -- $sedit(quoteDefault) {
 	always	{ set sedit($t,quote) 1 }
 	never	{ set sedit($t,quote) 0 }
@@ -192,7 +210,7 @@ proc Sedit_Start { draft } {
 	close $in
 	SeditPositionCursor $t
     }
-    focus $t
+    SeditSetIsigHeaders $t
     SeditMimeParse $t
     if {$sedit(iso)} {
 	SeditInitMimeType $draft $t
@@ -282,11 +300,15 @@ or do nothing?"
 	Widget_Frame $f f Dialog
 	$f.f configure -bd 10
 	Widget_AddBut $f.f ok "Abort" [list SeditAbortDirect $draft $t]
-	Widget_AddBut $f.f send "Send" [list SeditSend $draft $t]
-	Widget_AddBut $f.f save "Save" \
+	Widget_AddBut $f.f send "Send\n(Ctrl-c)" [list SeditSend $draft $t]
+	Widget_AddBut $f.f save "Save\n(Ctrl-s)" \
 		[list SeditSave $draft $t SeditNuke]
-	Widget_AddBut $f.f no "Do nothing" [list destroy $f]
+	Widget_AddBut $f.f no "Do nothing\n(Return)" [list destroy $f]
+	bind $f.f <Return>    "$f.f.no   flash ; $f.f.no   invoke"
+	bind $f.f <Control-c> "$f.f.send  flash ; $f.f.send  invoke"
+	bind $f.f <Control-s> "$f.f.save flash ; $f.f.save invoke"
 	Widget_PlaceDialog $t $f
+	focus $f.f
     } else {
 	SeditNuke $draft $t
     }
@@ -301,12 +323,6 @@ proc SeditAbortDirect { draft t } {
 }
 proc SeditAbort { draft t } {
     global sedit
-    if {0} {
-	if ![SeditIsDirty $t] {
-	    SeditAbortDirect $draft $t
-	    return
-	}
-    }
     if [catch {frame $t.abort -bd 4 -relief ridge -class Dialog} f] {
 	# dialog already up
 	SeditAbortConfirm $t.abort $t abort
@@ -320,15 +336,19 @@ You might prefer Save&Quit."
     frame $f.but -bd 10 -relief flat
     pack $f.but -expand true -fill both
     set sedit($t,abort) nop
-    Widget_AddBut $f.but ok "Abort" [list SeditAbortConfirm $f $t abort] left
-    Widget_AddBut $f.but save "Save&Quit" [list SeditAbortConfirm $f $t save] left
-    Widget_AddBut $f.but nop "Do Nothing" [list SeditAbortConfirm $f $t nop] right
+    Widget_AddBut $f.but ok "Abort\n(Return)" [list SeditAbortConfirm $f $t abort] left
+    Widget_AddBut $f.but save "Save&Quit\n(Ctrl-s)" [list SeditAbortConfirm $f $t save] left
+    Widget_AddBut $f.but nop "Do Nothing\n(Ctrl-c)" [list SeditAbortConfirm $f $t nop] right
+    bind $f.but <Return>    "$f.but.ok   flash ; $f.but.ok   invoke"
+    bind $f.but <Control-c> "$f.but.nop  flash ; $f.but.nop  invoke"
+    bind $f.but <Control-s> "$f.but.save flash ; $f.but.save invoke"
     Widget_PlaceDialog $t $f
+    focus $f.but
     tkwait window $f
     switch $sedit($t,abort) {
 	abort {SeditAbortDirect $draft $t}
 	save  {SeditSave $draft $t SeditNuke}
-	default { # nothing }
+	default { focus $t }
     }
 }
 proc SeditAbortConfirm { f t yes } {
@@ -343,6 +363,7 @@ proc SeditNuke { draft t } {
     catch {destroy $t.seditDirty}
     catch {destroy $sedit($t,toplevel).whom}
     catch {destroy $sedit($t,toplevel).spell}
+    update idletasks
     Exwin_Dismiss $sedit($t,toplevel)
 }
 proc SeditMsg { t text } {
@@ -356,18 +377,23 @@ proc SeditMsg { t text } {
 }
 
 proc SeditSend { draft t } {
-    global sedit exmh
+    global sedit exmh intelligentSign editor
     set id [SeditId $draft]
+    SeditCheckForIsigHeaders $t
     Exmh_Debug SeditSend id=$id action=$exmh($id,action)
     if {$sedit(autoSign) && ($sedit($t,sigfile) != "") &&
 	([string compare $exmh($id,action) "dist"] != 0)} {
 	set b .sedit${id}.but
 	set cmd [option get $b.sign command {}]
 	if {[string length $cmd] == 0} {
-	    Exmh_Debug SeditSend null cmd for $b.sign"
+	    Exmh_Debug SeditSend null cmd for $b.sign
 	    set cmd {SeditSign $draft $t}
 	}
-	eval $cmd $sedit($t,sigfile)
+	if {[string compare $sedit($t,sigfile) {///auto///}] == 0} {
+	    SeditSignIntelligent $draft $t [SeditProperSigfileDefault]
+	} else {
+	    eval $cmd $sedit($t,sigfile)
+	}
     }
     foreach cmd [info commands Hook_SeditSave*] {
 	if [catch {$cmd $draft $t} err] {
@@ -380,14 +406,14 @@ proc SeditSend { draft t } {
     if {$sedit(iso)} {
 	SeditFixupCharset $draft $t
     }
-    if [SeditSave $draft $t] {
+    if [SeditSave $draft $t {} 0] {
 	global env sedit
 	$sedit($t,toplevel).but.send config -state disabled
 	# Decide if this file needs to go through mhn
 	if {$sedit($t,mhn) && ![catch {exec grep -l ^# $draft}]} {
 	    set env(mhdraft) $draft
 	    SeditMsg $t "Running mhn..."
-	    if [catch {exec mhn $draft} err] {
+	    if [catch {exec $editor(mhn) $draft} err] {
 		SeditMsg $t $err
 		$sedit($t,toplevel).but.send config -state normal
 		return
@@ -433,19 +459,40 @@ proc SeditSend { draft t } {
 	    if {[string compare $exmh(folder) $mhProfile(draft-folder)] == 0} {
 		Scan_Folder $exmh(folder)
 	    }
+	    SeditMsg $t "Message saved and sent $time"
 	}
     }
 }
 
-proc SeditSave { draft t {hook {}} } {
-    global sedit mhProfile
+proc SeditSave { draft t {hook {}} {isigw 1} } {
+    global sedit mhProfile exmh
     if [catch {
 	SeditMsg $t "Saving message..."
 	set out [open $draft w]
-	if {$sedit($t,format)} {
-	    SeditFormatMail $t $out
+	if {([string compare $sedit($t,format) "Never"] != 0)} {
+	    SeditFormatMail $t $out $isigw
 	} else {
-	    puts $out [$t get 1.0 end]
+	    # Prevent duplicate X-Mailer or X-Exmh-Isig-* headers
+	    set id $sedit($t,id)
+	    SeditCheckForIsigHeaders $t
+	    if {[catch {set end [$t index hlimit]}] &&
+		[catch {set end [$t index header]}]} {
+		    set end end
+	    }
+	    set X1 [$t get 1.0 $end]
+	    set X2 [$t get $end end]
+	    regsub -all -nocase "(^|\n)(x-mailer:\[^\n\]*\n)+" $X1 {\1} X1
+	    regsub -all -nocase "(^|\n)(x-exmh-isig-(comptype|folder):\[^\n\]*\n)+" $X1 {\1} X1
+	    # No X-Mailer on redistributed messages
+	    if {[string compare $exmh($id,action) dist] != 0} {
+		puts $out "X-Mailer: exmh $exmh(version)"
+	    }
+	    # Replace X-Exmh-Isig-* headers if necessary
+	    if {$isigw} {
+		puts $out "X-Exmh-Isig-CompType: $sedit($t,isigc)"
+		puts $out "X-Exmh-Isig-Folder: $sedit($t,isigf)"
+	    }
+	    puts -nonewline $out "$X1$X2"
 	}
 	close $out
 	SeditMsg $t "Message saved"
@@ -470,16 +517,18 @@ proc SeditAlternate { draft t } {
     Edit_Done alternate [SeditId $draft]
 }
 proc SeditSaveBody { t outfile } {
-    set out [open $outfile w]
-    puts $out [$t get [$t index "header + 1 line"] end]
+    set out [open $outfile w 0600]
+    puts -nonewline $out [$t get [$t index "header + 1 line"] end]
     close $out
 }
 
 proc SeditReplaceBody { t infile } {
     set in [open $infile]
+    set tags [$t tag names "header + 1 line"]
     $t delete "header + 1 line" end
-    $t insert end [read $in]
+    $t insert end [read $in] $tags
     close $in
+    SeditMimeParse $t	;# Reconstruct formatting state
 }
 
 proc SeditMarkSent { t } {
@@ -505,11 +554,18 @@ proc Sedit_CheckPoint {} {
 	}
     }
 }
+proc SeditPeriodicSave {} {
+    global sedit
+    Sedit_CheckPoint
+    set sedit(checkpoint) [after [expr 1000 * 60] SeditPeriodicSave]
+}
+
 proc SeditFixupMhn { draft t } {
     global sedit
     set state header
     set mhn 0
     set lines {}
+    Exmh_Debug SeditFixupMhn
     for {set i 1} {[$t compare $i.0 < end]} {incr i} {
 	set line [$t get $i.0 $i.end]
 	set len [string length $line]
@@ -530,7 +586,7 @@ proc SeditFixupMhn { draft t } {
 	if [llength $lines] {
 	    SeditMsg $t "Cleaning up for MHN"
 	}
-	foreach i [lsort -decreasing $lines] {
+	foreach i [lsort -integer -decreasing $lines] {
 	    $t delete $i.0 "$i.end +1 char"
 	}
 	set sedit($t,8bit) 0	;# Let MHN do quote-printable
@@ -572,3 +628,43 @@ proc SeditSetHeader { t key value } {
 	$t tag remove Charset $L.0 $L.end
     }
 }
+
+# Set/Replace/Append a message header
+# (does not handle duplicate headers)
+proc SeditSetHeader { t key value {append NO}} {
+    if [string match NO $append] {
+	unset append
+    }
+
+    # let the text widget search for the header
+    set start [$t search -nocase -regexp ^${key}: 1.0 header]
+
+    if {[string compare "" $start]} {
+	# find the end of the header, looking past any continuation lines
+	set end [$t search -regexp {^[^	 ]} "$start + 1 line" header]
+
+	# in case user mangled the header.
+        if {[string match "" $end]} {
+            set end "$start +1 line"
+        }
+
+	# if we are appending, do it now.
+	if [info exists append] {
+	    $t insert "$end -1c" "$append$value"
+	    return
+	}
+
+	# otherwise, delete the whole header to set up for insertion.
+	$t delete $start $end
+    } else {
+	# insert a new header near the end of the headers
+	set start "header linestart"
+    }
+
+    # insert the new/replaced header
+    $t insert $start "[string toupper [string index $key 0]][string tolower [string range $key 1 end]]: $value\n"
+    $t tag remove Charset $start "$start lineend"
+}
+
+
+
